#ifndef ___CONVERT_HPP___
#define ___CONVERT_HPP___

#include "Numeric.hpp"
#include "Pixel.hpp"

struct affine
{
	const matrix3x3 mat;

	affine( const matrix3x3 & m )
		: mat(inverse(m))
	{}

	template< typename Image, typename Position >
	typename Image::value_type operator()(
			const Image & img, 
			Position x,
			Position y
			) const
	{
		vec3 p; p[0] = x; p[1] = y; p[2] = 1.0;
		vec3 v = prod( p, mat );
		return img( (float)v[0]/v[2], (float)v[1]/v[2] );
	}
};

struct bilinear
{
	template< typename Image, typename Position >
	typename Image::value_type operator()(
			const Image & img, 
			Position x,
			Position y
			) const
	{
		int u = x;
		int v = y;
		float a = x - u;
		float b = y - v;
		
		return   (1-a) * (1-b) * img( u   , v )
		       + a     * (1-b) * img( u+1 , v )
			   + (1-a) * b     * img( u   , v+1 )
			   + a     * b     * img( u+1 , v+1 );
	}
};

template< typename value, int N >
struct filter
{
	value f[N][N];
	value high;
	value low;
	value range;
	
	filter( value _f[N][N] )
	{
		memcpy( f, _f, sizeof(f) );

		high = low = 0;

		for( int i = 0; i < N*N; ++i )
		{
			value w = f[i/N][i%N];

			if( w > 0 )
				    high += w;
			else	low  += w;
		}

		range = high - low;
	}

	template< typename Image, typename Position >
	typename Image::value_type operator()(
			const Image & img,
			Position x,
			Position y
			) const
	{
		typedef typename Image::value_type pixel;
		typedef Container<pixel>::type container;
		
		container sum = 0;

		for( int i = 0; i < N; ++i )
			for( int j = 0; j < N; ++j )
			{
				Position tx = x+j-N/2;
				Position ty = y+i-N/2;
				sum += img(tx,ty) * f[i][j];
			}
		
		const container min_value = Range<pixel>::MAX * low;
		const container src_range = Range<pixel>::MAX * range;
		const pixel     dst_range = Range<pixel>::MAX - Range<pixel>::MIN;

		return (sum - min_value) * dst_range / src_range + Range<pixel>::MIN;
	}
};

template< typename T, int N >
filter<T,N> Filter( T f[][N] )
{
	return filter<T,N>(f);
}

template< typename Matrix >
struct filter2
{
	typedef typename Matrix::value_type  value_type;

	const Matrix & f;
	const int width;
	const int height;
	value_type high;
	value_type low;
	value_type range;
	
	filter2( const Matrix & _f, int _width, int _height )
		: f(_f), width(_width), height(_height)
	{
		high = low = 0;

		for( int i = 0; i < width*height; ++i )
		{
			value_type w = f(i%width,i/width);

			if( w > 0 )
				    high += w;
			else	low  += w;
		}

		range = high - low;
	}

	template< typename Image, typename Position >
	typename Image::value_type operator()(
			const Image & img,
			Position x,
			Position y
			) const
	{
		typedef typename Image::value_type pixel;
		typedef Container<pixel>::type container;
		
		container sum = 0;

		for( int i = 0; i < height; ++i )
			for( int j = 0; j < width; ++j )
			{
				Position tx = x+j - width/2;
				Position ty = y+i - height/2;
				sum += img(tx,ty) * f(j,i);
			}
		
		const container min_value = Range<pixel>::MAX * low;
		const container src_range = Range<pixel>::MAX * range;
		const pixel     dst_range = Range<pixel>::MAX - Range<pixel>::MIN;

		return (sum - min_value) * dst_range / src_range + Range<pixel>::MIN;
	}
};

template< typename Matrix >
filter2<Matrix> Filter2( const Matrix & m, int w, int h )
{
	return filter2<Matrix>(m,w,h);
}

template< typename Pixel >
struct binary
{
	const Pixel border;

	binary( const Pixel & _border )
		: border(_border)
	{}

	template< typename Image, typename Position >
	typename Image::value_type operator()(
			const Image & img,
			Position x,
			Position y
			) const
	{
		typedef typename Image::value_type pixel;
		return img(x,y) > border ? Range<pixel>::MAX : Range<pixel>::MIN;
	}
};

template< typename Pixel >
binary<Pixel> Binary( const Pixel & border )
{
	return binary<Pixel>(border);
}

template< typename Pixel >
struct paint
{
	const Pixel & fill;

	paint( const Pixel & _fill )
		: fill(_fill)
	{}

	template< typename Image, typename Position >
	typename Image::value_type operator()(
		const Image & img,
		const Position & x,
		const Position & y
		) const
	{
		return fill;
	}
};

template< typename Pixel >
paint<Pixel> Paint( const Pixel & fill )
{
	return paint<Pixel>(fill);
}

struct biCubic
{
	static float fc( float v )
	{
		v = abs(v);

		if( v < 1.0f )
			return 1.0f - 2.0f*v*v + v*v*v;

		if( v < 2.0f )
			return 4.0f - 8.0f*v + 5.0f*v*v - v*v*v;

		return 0.0f;
	}

	template< typename Image, typename Position >
	typename Image::value_type operator()(
		const Image & img,
		const Position & x,
		const Position & y
		) const
	{
		int   u = x,   v = y;
		float a = x-u, b = y-v;

		float sum = 0.0f;

		for( int py = -1; py <= 2; ++py )
		{
			float tsum = 0.0f;

			for( int px = -1; px <= 2; ++px )
				tsum += fc(-a+px) * img( u+px, v+py );

			sum += fc(-b+py) * tsum;
		}

		return round((typename Image::value_type)sum);
	}
};

inline
ublas::matrix<float> Gaussian( const int w, const int h )
{
	const float s = 1.0f;
	ublas::matrix<float> m(w,h);

	for( int i = 0; i < w*h; ++i )
	{
		float x = w == 1 ? 0.0f : ((float)(i%w)-w/2) / (w/2);
		float y = h == 1 ? 0.0f : ((float)(i/w)-h/2) / (h/2);
		m(i%w,i/w) = exp((-x*x-y*y)/(s*2.0f));
	}

	return m;
}

inline 
ublas::matrix<float> Smooth( const int n )
{
	ublas::matrix<float> m(n,n);
	for( int i = 0; i < n*n; ++i )
		m(i%n,i/n) = 1.0f;
	
	return m;
}

template< unsigned int N >
struct median
{
	template< typename Image >
	typename Image::value_type operator()(
			const Image & img,
			int x, int y
		) const
	{
		typedef typename Image::value_type value_type;
		value_type p[N*N];

		for( int i = 0; i < N*N; ++i )
			p[i] = img( (i%N)-N/2+x, (i/N)-N/2+y );

		std::sort( p, p+N*N );
		return p[N/2];
	}
};

namespace detail
{
	int diff1[3][3] =
	{
		{ 1,  0, -1},
		{ 1,  0, -1},
		{ 1,  0, -1}
	};

	int diff2[3][3] =
	{
		{ 1,  1,  1},
		{ 0,  0,  0},
		{-1, -1, -1}
	};
}

struct edge
{
	filter<int,3> f1;
	filter<int,3> f2;

	edge() : f1(detail::diff1), f2(detail::diff2)
	{}

	template< typename Image >
	typename Image::value_type operator()(
			const Image & img,
			int tx, int ty
		) const
	{
		typedef typename Image::value_type value_type;
		typedef Container<value_type>::type container;
		
		float h = (Range<value_type>::MAX - Range<value_type>::MIN) / 2.0f;
		float x = f1(img,tx,ty) - h;
		float y = f2(img,tx,ty) - h;
		
		return round<unsigned char>( sqrt(x*x+y*y)/sqrt(h*h+h*h)*Range<value_type>::MAX );
	}
};

#endif
