#ifndef ___IMAGE_HPP___
#define ___IMAGE_HPP___

#include <boost/shared_ptr.hpp>
#include "Numeric.hpp"
#include "Pixel.hpp"

template< typename left, typename right >
class expression
{
	const left  & lhs;
	const right & rhs;

public:

	typedef typename left::value_type value_type;

	expression( const left & l, const right & r )
		: lhs(l), rhs(r)
	{
	}

	template< typename Position >
	value_type operator() (
			Position x,
			Position y
		) const
	{
		return rhs( lhs, x, y );
	}
};

#define SPECIALIZE_EXPRESSION( RIGHT_TYPE, VALUE_TYPE )	\
	template< typename left >			\
	class expression< left, RIGHT_TYPE >\
	{									\
		const left  & lhs;				\
		const RIGHT_TYPE  & rhs;		\
										\
	public:								\
										\
		typedef VALUE_TYPE value_type;	\
										\
		expression(						\
			const left & l,				\
			const RIGHT_TYPE & r		\
			)							\
			: lhs(l), rhs(r){}			\
										\
		template< typename Position >	\
		VALUE_TYPE operator() (			\
				Position x,				\
				Position y				\
			) const						\
		{								\
			return rhs( lhs, x, y );	\
		}								\
	};

template< typename left, typename right >
expression<left,right> Expression( const left & l, const right & r )
{
	return expression<left,right>(l,r);
}


#define DEFINE_STD_PIXEL_TYPE( PIXEL )								\
	typedef PIXEL							value_type;			\
	typedef PIXEL &							reference;			\
	typedef const PIXEL &					const_reference;	\
	typedef Range<PIXEL>					range;				\
	typedef typename Container<PIXEL>::type	container_type;


template< typename Pixel >
class IImage
{
	public:
		
		DEFINE_STD_PIXEL_TYPE(Pixel)

		IImage( int w, int h )
			: width(w), height(h)
		{}

		const int	width;
		const int	height;

		bool inside( int x, int y ) const
		{
			return !( x < 0 || width <= x || y < 0 || height <= y );
		}

		virtual reference       operator()( int x, int y ) = 0;
		virtual const_reference operator()( int x, int y ) const = 0;
};

template< typename left1, typename left2, typename right >
expression<expression<left1,left2>,right>
	operator + (
		const expression<left1,left2> &  l,
		const right & r
	)
{
	return expression<expression<left1,left2>,right>(l,r);
}

template< typename Pixel, typename right >
expression<IImage<Pixel>,right>
	operator + (
		const IImage<Pixel> & image,
		const right & r
		)
{
	return expression<IImage<Pixel>,right>(image,r);
}

template< typename src, typename dst >
void convert( const src & s, dst & d, int width, int height )
{
	for( int y = 0; y < height; ++y )
		for( int x = 0; x < width; ++x )
			d(x,y) = s(x,y);
	
	return;
}

class GrayImage : public IImage<unsigned char>
{
private:

	std::vector<unsigned char>	data;

	int offset( int x, int y ) const
	{
		return y*width + x;
	}

	void revision( int & x, int & y ) const
	{
		x = std::max( 0, std::min(width-1,  x) );
		y = std::max( 0, std::min(height-1, y) );
	}

public:

	GrayImage( int _width, int _height )
		: IImage(_width,_height),
		  data(_width*_height,0)
	{
	}

	reference operator()( int x, int y )
	{
		revision(x,y);
		return data[offset(x,y)];
	}

	const_reference operator()( int x, int y ) const
	{
		revision(x,y);
		return data[offset(x,y)];
	}

	template< typename Image >
	GrayImage & operator = ( const Image & img )
	{
		convert( img, *this, width, height );
		return *this;
	}
};

typedef boost::shared_ptr<GrayImage>	GrayImagePtr;

//CɂŐVASYT@F
template< typename Image >
void drawLine(
		Image & image,
		int x1, int y1, int x2, int y2,
		const typename Image::value_type & color
		)
{
    int dx, dy, s, step;

    dx = abs(x2 - x1);  dy = abs(y2 - y1);
    if (dx > dy) {
        if (x1 > x2) {
            step = (y1 > y2) ? 1 : -1;
            s = x1;  x1 = x2;  x2 = s;  y1 = y2;
        } else step = (y1 < y2) ? 1: -1;
       image(x1, y1) = color;
        s = dx >> 1;
        while (++x1 <= x2) {
            if ((s -= dy) < 0) {
                s += dx;  y1 += step;
            };
            image(x1, y1) = color;
        }
    } else {
        if (y1 > y2) {
            step = (x1 > x2) ? 1 : -1;
            s = y1;  y1 = y2;  y2 = s;  x1 = x2;
        } else step = (x1 < x2) ? 1 : -1;
        image(x1, y1) = color;
        s = dy >> 1;
        while (++y1 <= y2) {
            if ((s -= dx) < 0) {
                s += dy;  x1 += step;
            }
            image(x1, y1) = color;
        }
    }
}

#endif
